import os
from io import BytesIO
from unittest import TestCase as BaseTestCase

import fs
import requests
from google.api_core.exceptions import Conflict, NotFound

from gcloud_storage_emulator.server import create_server
from gcloud_storage_emulator.settings import STORAGE_BASE, STORAGE_DIR


def _get_storage_client(http):
    """Gets a python storage client"""
    os.environ["STORAGE_EMULATOR_HOST"] = "http://localhost:9023"

    # Cloud storage uses environment variables to configure api endpoints for
    # file upload - which is read at module import time
    from google.cloud import storage
    return storage.Client(
        project="[PROJECT]",
        _http=http,
        client_options={"api_endpoint": "http://localhost:9023"},
    )


class ServerBaseCase(BaseTestCase):

    @classmethod
    def setUpClass(cls):
        cls._server = create_server("localhost", 9023, in_memory=False)
        cls._server.start()

    @classmethod
    def tearDownClass(cls):
        cls._server.stop()

    def setUp(self):
        self._session = requests.Session()
        self._client = _get_storage_client(self._session)
        ObjectsTests._server.wipe()


class BucketsTests(BaseTestCase):

    @classmethod
    def setUpClass(cls):
        cls._server = create_server("localhost", 9023, in_memory=False)
        cls._server.start()

    @classmethod
    def tearDownClass(cls):
        cls._server.stop()

    def setUp(self):
        BucketsTests._server.wipe()
        self._session = requests.Session()
        self._client = _get_storage_client(self._session)

    def test_bucket_creation(self):
        bucket = self._client.create_bucket("bucket_name")
        self.assertEqual(bucket.project_number, 1234)

    def test_bucket_creation_no_override(self):
        self._client.create_bucket("bucket_name")
        with self.assertRaises(Conflict):
            self._client.create_bucket("bucket_name")

    def test_bucket_list(self):
        bucket = self._client.create_bucket("bucket_name")
        all_bucket_names = [b.name for b in self._client.list_buckets()]
        self.assertIn(bucket.name, all_bucket_names)

    def test_bucket_get_existing(self):
        bucket = self._client.create_bucket("bucket_name")
        fetched_bucket = self._client.get_bucket("bucket_name")
        self.assertEqual(fetched_bucket.name, bucket.name)

    def test_bucket_get_existing_with_dot(self):
        bucket = self._client.create_bucket("bucket.name")
        fetched_bucket = self._client.get_bucket("bucket.name")
        self.assertEqual(fetched_bucket.name, bucket.name)

    def test_bucket_get_non_existing(self):
        with self.assertRaises(NotFound):
            self._client.get_bucket("bucket_name")

    def test_bucket_delete(self):
        bucket = self._client.create_bucket("bucket_name")
        bucket.delete()

        with self.assertRaises(NotFound):
            self._client.get_bucket("bucket_name")

    def test_bucket_delete_removes_file(self):
        bucket = self._client.create_bucket("bucket_name")
        bucket.delete()

        with fs.open_fs(STORAGE_BASE + STORAGE_DIR) as pwd:
            self.assertFalse(pwd.exists("bucket_name"))

    def test_bucket_delete_non_existing(self):
        # client.bucket doesn't create the actual bucket resource remotely,
        # it only instantiate it in the local client
        bucket = self._client.bucket("bucket_name")
        with self.assertRaises(NotFound):
            bucket.delete()

    def test_bucket_delete_non_empty(self):
        bucket = self._client.create_bucket("bucket_name")
        blob = bucket.blob("canttouchme.txt")
        blob.upload_from_string("This should prevent deletion if not force")

        with self.assertRaises(Conflict):
            bucket.delete()

        blob = bucket.get_blob("canttouchme.txt")
        self.assertIsNotNone(blob)

    def test_bucket_force_delete(self):
        bucket = self._client.create_bucket("bucket_name")
        blob = bucket.blob("cantouchme.txt")
        blob.upload_from_string("This should prevent deletion if not force")

        bucket.delete(force=True)

        blob = bucket.get_blob("cantouchme.txt")
        self.assertIsNone(blob)

        with fs.open_fs(STORAGE_BASE + STORAGE_DIR) as pwd:
            self.assertFalse(pwd.exists('bucket_name'))
    # TODO: test delete-force


class DefaultBucketTests(BaseTestCase):
    def tearDown(self):
        if self._server:
            self._server.stop()
        return super().tearDown()

    def test_bucket_created(self):
        self._server = create_server("localhost", 9023, in_memory=True, default_bucket="example.appspot.com")
        self._server.start()
        self._session = requests.Session()
        self._client = _get_storage_client(self._session)
        self._client.get_bucket("example.appspot.com")


class ObjectsTests(ServerBaseCase):

    def test_upload_from_string(self):
        content = "this is the content of the file\n"
        bucket = self._client.create_bucket("testbucket")
        blob = bucket.blob("testblob-name.txt")
        blob.upload_from_string(content)

        with fs.open_fs(STORAGE_BASE + STORAGE_DIR) as pwd:
            read_content = pwd.readtext("testbucket/testblob-name.txt")
            self.assertEqual(read_content, content)

    def test_upload_from_text_file(self):
        text_test = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'test_text.txt')
        bucket = self._client.create_bucket("testbucket")
        blob = bucket.blob("test_text.txt")
        with open(text_test, "rb") as file:
            blob.upload_from_file(file)

            with fs.open_fs(STORAGE_BASE + STORAGE_DIR) as pwd:
                read_content = pwd.readtext("testbucket/test_text.txt")

        with open(text_test, "rb") as file:
            expected_content = str(file.read(), encoding="utf-8")
            self.assertEqual(read_content, expected_content)

    def test_upload_from_bin_file(self):
        test_binary = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'test_binary.png')
        bucket = self._client.create_bucket("testbucket")
        blob = bucket.blob("binary.png")
        with open(test_binary, "rb") as file:
            blob.upload_from_file(file)

        with fs.open_fs(STORAGE_BASE + STORAGE_DIR) as pwd:
            read_content = pwd.readbytes("testbucket/binary.png")

        with open(test_binary, "rb") as file:
            expected_content = file.read()
            self.assertEqual(read_content, expected_content)

    def test_upload_from_bin_file_cr_lf(self):
        content = b'\r\rheeeeei\r\n'
        test_binary = BytesIO(content)
        bucket = self._client.create_bucket("testbucket")
        blob = bucket.blob("binary_cr.png")

        blob.upload_from_file(
            test_binary, size=len(content)
        )

        with fs.open_fs(STORAGE_BASE + STORAGE_DIR) as pwd:
            read_content = pwd.readbytes("testbucket/binary_cr.png")

        self.assertEqual(read_content, content)

    def test_get(self):
        file_name = "testblob-name.txt"
        content = "this is the content of the file\n"
        bucket = self._client.create_bucket("testbucket")
        blob = bucket.blob(file_name)
        blob.upload_from_string(content)

        blob = bucket.get_blob(file_name)
        self.assertEqual(blob.name, file_name)

    def test_get_unicode(self):
        file_name = "tmp.ąćęłńóśźż.马铃薯.zip"
        content = "this is the content of the file\n"
        bucket = self._client.create_bucket("testbucket")
        blob = bucket.blob(file_name)
        blob.upload_from_string(content)

        blob = bucket.get_blob(file_name)
        self.assertEqual(blob.name, file_name)

    def test_get_nonexistant(self):
        bucket = self._client.create_bucket("testbucket")
        res = bucket.get_blob("idonotexist")

        self.assertIsNone(res)

        blob = bucket.blob("iexist")
        blob.upload_from_string("some_fake_content")
        blob = bucket.get_blob("idonotexist")

        self.assertIsNone(res)

    def test_download_as_string(self):
        content = "The quick brown fox jumps over the lazy dog\n"
        bucket = self._client.create_bucket("testbucket")

        blob = bucket.blob("iexist")
        blob.upload_from_string(content)

        blob = bucket.get_blob("iexist")
        fetched_content = blob.download_as_string()
        self.assertEqual(fetched_content, content.encode('utf-8'))

    def test_download_binary_to_file(self):
        test_binary = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'test_binary.png')
        bucket = self._client.create_bucket("testbucket")

        blob = bucket.blob("binary.png")
        with open(test_binary, "rb") as file:
            blob.upload_from_file(file, content_type="image/png")

        blob = bucket.get_blob("binary.png")
        fetched_file = BytesIO()
        blob.download_to_file(fetched_file)

        with open(test_binary, "rb") as file:
            self.assertEqual(fetched_file.getvalue(), file.read())

    def test_download_text_to_file(self):
        test_text = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'test_text.txt')
        bucket = self._client.create_bucket("testbucket")

        blob = bucket.blob("text.txt")
        with open(test_text, "rb") as file:
            blob.upload_from_file(file, content_type="text/plain; charset=utf-8")

        blob = bucket.get_blob("text.txt")
        fetched_file = BytesIO()
        blob.download_to_file(fetched_file)

        with open(test_text, "rb") as file:
            self.assertEqual(fetched_file.getvalue(), file.read())

    def test_delete_object(self):
        bucket = self._client.create_bucket("bucket_name")
        blob = bucket.blob("canttouchme.txt")
        blob.upload_from_string("File content")

        with fs.open_fs(STORAGE_BASE + STORAGE_DIR) as pwd:
            self.assertTrue(pwd.exists("bucket_name/canttouchme.txt"))
            blob.delete()

            self.assertIsNone(bucket.get_blob("cantouchme.txt"))
            self.assertFalse(pwd.exists("bucket_name/canttouchme.txt"))

    def test_create_within_directory(self):
        bucket = self._client.create_bucket("bucket_name")
        blob = bucket.blob("this/is/a/nested/file.txt")
        blob.upload_from_string("Not even joking!")

        with fs.open_fs(STORAGE_BASE + STORAGE_DIR) as pwd:
            read_content = pwd.readtext("bucket_name/this/is/a/nested/file.txt")
            self.assertEqual(read_content, "Not even joking!")

    def test_create_within_multiple_time_does_not_break(self):
        bucket = self._client.create_bucket("bucket_name")
        blob = bucket.blob("this/is/a/nested/file.txt")
        blob.upload_from_string("Not even joking!")

        bucket.blob("this/is/another/nested/file.txt")
        blob.upload_from_string("Yet another one")

        with fs.open_fs(STORAGE_BASE + STORAGE_DIR) as pwd:
            self.assertTrue(pwd.exists("bucket_name/this/is/a/nested/file.txt"))

    def _assert_blob_list(self, expected, actual):
        self.assertEqual(
            [b.name for b in expected],
            [b.name for b in actual])

    def test_list_blobs_on_nonexistent_bucket(self):
        blobs = self._client.list_blobs("bucket_name")
        with self.assertRaises(NotFound):
            list(blobs)

    def test_list_blobs_on_empty_bucket(self):
        bucket = self._client.create_bucket("bucket_name")
        blobs = self._client.list_blobs(bucket)
        self._assert_blob_list(blobs, [])

    def test_list_blobs_on_entire_bucket(self):

        bucket_1 = self._client.create_bucket("bucket_name_1")
        bucket_2 = self._client.create_bucket("bucket_name_2")

        blob_1 = bucket_1.blob("a/b.txt")
        blob_1.upload_from_string("text")

        blob_2 = bucket_1.blob("c/d.txt")
        blob_2.upload_from_string("text")

        blob_3 = bucket_2.blob("a/b.txt")
        blob_3.upload_from_string("text")

        blobs = self._client.list_blobs(bucket_1)
        self._assert_blob_list(blobs, [blob_1, blob_2])

    def test_list_blobs_with_prefix(self):
        bucket = self._client.create_bucket("bucket_name")

        blob_1 = bucket.blob("a/b.txt")
        blob_1.upload_from_string("text")

        blob_2 = bucket.blob("a/b/c.txt")
        blob_2.upload_from_string("text")

        blob_3 = bucket.blob("b/c.txt")
        blob_3.upload_from_string("text")

        blobs = self._client.list_blobs(bucket, prefix='a')

        self._assert_blob_list(blobs, [blob_1, blob_2])

    def test_list_blobs_with_prefix_and_delimiter(self):
        bucket = self._client.create_bucket("bucket_name")

        blob_1 = bucket.blob("a/b.txt")
        blob_1.upload_from_string("text")

        blob_2 = bucket.blob("a/c.txt")
        blob_2.upload_from_string("text")

        blob_3 = bucket.blob("a/b/c.txt")
        blob_3.upload_from_string("text")

        blob_4 = bucket.blob("b/c.txt")
        blob_4.upload_from_string("text")

        blobs = self._client.list_blobs(bucket, prefix='a', delimiter='/')

        self._assert_blob_list(blobs, [blob_1, blob_2])

    def test_bucket_copy_existing(self):
        bucket = self._client.create_bucket("bucket_name")

        blob_1 = bucket.blob("a/b.txt")
        blob_1.upload_from_string("text")

        blob_2 = bucket.rename_blob(blob_1, "c/d.txt")

        blobs = self._client.list_blobs(bucket)
        self._assert_blob_list(blobs, [blob_2])

    def test_bucket_copy_non_existing(self):
        bucket = self._client.create_bucket("bucket_name")

        blob_1 = bucket.blob("a/b.txt")

        with self.assertRaises(NotFound):
            bucket.rename_blob(blob_1, "c/d.txt")


class HttpEndpointsTest(ServerBaseCase):
    """ Tests for the HTTP endpoints defined by server.HANDLERS. """

    def _url(self, path):
        return os.environ["STORAGE_EMULATOR_HOST"] + path

    def test_download_by_url(self):
        """ Objects should be downloadable over HTTP from the emulator client. """
        content = "Here is some content"
        bucket = self._client.create_bucket("anotherbucket")
        blob = bucket.blob("something.txt")
        blob.upload_from_string(content)

        url = self._url("/anotherbucket/something.txt")
        response = requests.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, content.encode('utf-8'))

    def test_download_file_within_folder(self):
        """ Cloud Storage allows folders within buckets, so the download URL should allow for this.
        """
        content = "Here is some content"
        bucket = self._client.create_bucket("yetanotherbucket")
        blob = bucket.blob("folder/contain~ing/something~v-1.0.α.txt")
        blob.upload_from_string(content)

        url = self._url("/yetanotherbucket/folder/contain~ing/something~v-1.0.α.txt")
        response = requests.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, content.encode('utf-8'))

    def test_wipe(self):
        """ It should wipe the data """
        storage_path = STORAGE_BASE + STORAGE_DIR
        content = "Here is some content"
        bucket = self._client.create_bucket("anotherbucket1")
        blob = bucket.blob("something.txt")
        blob.upload_from_string(content)

        url = self._url("/wipe")
        response = requests.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(len(os.listdir(storage_path)) == 0)

    def test_wipe_keep_buckets(self):
        """It should wipe the data but keep the root buckets"""
        blob_path = "something.txt"
        bucket_name = "anewone"
        content = "Here is some content"
        bucket = self._client.create_bucket(bucket_name)
        blob = bucket.blob(blob_path)
        blob.upload_from_string(content)

        url = self._url("/wipe?keep-buckets=true")
        response = requests.get(url)
        self.assertEqual(response.status_code, 200)

        fetched_bucket = self._client.get_bucket(bucket_name)
        self.assertEqual(fetched_bucket.name, bucket.name)
        with self.assertRaises(NotFound):
            fetched_bucket.blob(blob_path).download_as_text()



import json
import os
from unittest import TestCase as BaseTestCase

from gcloud_storage_emulator.exceptions import NotFound
from gcloud_storage_emulator.settings import STORAGE_BASE, STORAGE_DIR
from gcloud_storage_emulator.storage import Storage


def _get_meta_path():
    return os.path.join(os.getcwd(), STORAGE_BASE, STORAGE_DIR, ".meta")


class StorageOSFSTests(BaseTestCase):
    def setUp(self):
        self.storage = Storage()
        self.storage.wipe()

    def tearDown(self):
        self.storage.wipe()

    def test_get_bucket_reads_from_meta(self):
        meta_path = _get_meta_path()
        buckets = {"key": "a"}

        with open(meta_path, "w") as file:
            json.dump({
                "buckets": buckets,
            }, file)

        # Force a re-read from file, this is usually done in the constructor
        self.storage._read_config_from_file()
        self.assertEqual(self.storage.get_bucket("key"), "a")

    def test_get_file_obj_reads_from_meta(self):
        meta_path = _get_meta_path()
        objects = {"key": {"inner_key": "a"}}

        with open(meta_path, "w") as file:
            json.dump({
                "objects": objects,
            }, file)

        # Force a re-read from file, this is usually done in the constructor
        self.storage._read_config_from_file()
        self.assertEqual(self.storage.get_file_obj("key", "inner_key"), "a")

    def test_get_file_obj_not_found(self):
        with self.assertRaises(NotFound):
            self.storage.get_file_obj("a_bucket", "a_file")

        self.storage.create_bucket("a_bucket", {})
        with self.assertRaises(NotFound):
            self.storage.get_file_obj("a_bucket", "a_file")

    def test_get_file_not_found(self):
        with self.assertRaises(NotFound):
            self.storage.get_file("a_bucket", "a_file")

        self.storage.create_bucket("a_bucket", {})
        with self.assertRaises(NotFound):
            self.storage.get_file("a_bucket", "a_file")

    def test_create_bucket_stores_meta(self):
        bucket_obj = {"key": "val"}
        self.storage.create_bucket("a_bucket", bucket_obj)

        meta_path = _get_meta_path()
        with open(meta_path, "r") as file:
            meta = json.load(file)
            self.assertEqual(meta["buckets"]["a_bucket"], bucket_obj)

    def test_create_file_stores_content(self):
        test_file = os.path.join(os.getcwd(), STORAGE_BASE, STORAGE_DIR, "a_bucket_name", "file_name.txt")
        content = "Łukas is a great developer".encode("utf8")
        file_obj = {}
        self.storage.create_file("a_bucket_name", "file_name.txt", content, file_obj)

        with open(test_file, "rb") as file:
            read_content = file.read()
            self.assertEqual(read_content, content)

    def test_create_file_stores_meta(self):
        content = "Łukas is a great developer".encode("utf8")
        file_obj = {"key": "val"}
        self.storage.create_file("a_bucket_name", "file_name.txt", content, file_obj)
        meta_path = _get_meta_path()
        with open(meta_path, "r") as file:
            meta = json.load(file)
            self.assertEqual(meta["objects"]["a_bucket_name"]["file_name.txt"], file_obj)

    def test_create_resumable_upload_stores_meta(self):
        file_obj = {"key": "val"}
        file_id = self.storage.create_resumable_upload("a_bucket_name", "file_name.png", file_obj)
        meta_path = _get_meta_path()
        with open(meta_path, "r") as file:
            meta = json.load(file)
            self.assertEqual(meta["resumable"][file_id], file_obj)

    def test_file_ids_dont_clash(self):
        file_obj = {"key": "val"}
        file_id_1 = self.storage.create_resumable_upload("a_bucket_name", "file_name.png", file_obj)
        file_id_2 = self.storage.create_resumable_upload("a_bucket_name", "file_name.png", file_obj)
        self.assertNotEqual(file_id_1, file_id_2)

    def test_create_file_for_resumable_upload(self):
        test_file = os.path.join(os.getcwd(), STORAGE_BASE, STORAGE_DIR, "a_bucket_name", "file_name.png")
        content = b"Randy is also a great developer"
        file_obj = {"bucket": "a_bucket_name", "name": "file_name.png"}
        file_id = self.storage.create_resumable_upload("a_bucket_name", "file_name.png", file_obj)
        self.storage.create_file_for_resumable_upload(file_id, content)

        with open(test_file, "rb") as file:
            read_content = file.read()
            self.assertEqual(read_content, content)

        with open(_get_meta_path(), "r") as file:
            meta = json.load(file)
            self.assertEqual(meta["objects"]["a_bucket_name"]["file_name.png"], file_obj)
            self.assertEqual(meta["resumable"], {})

        self.assertEqual(file_obj["size"], str(len(content)))

    def test_delete_bucket_stores_meta(self):
        bucket_obj = {"key": "val"}
        self.storage.create_bucket("a_bucket", bucket_obj)

        self.storage.delete_bucket("a_bucket")

        meta_path = _get_meta_path()
        with open(meta_path, "r") as file:
            meta = json.load(file)
            self.assertIsNone(meta["buckets"].get('a_bucket'))

    def test_wipe(self):
        bucket_a_obj = {"key_a": "val_a"}
        bucket_b_obj = {"key_b": "valb_"}
        self.storage.create_bucket("bucket_a", bucket_a_obj)
        self.storage.create_bucket("bucket_b", bucket_b_obj)
        self.storage.wipe()
        meta_path = _get_meta_path()
        self.assertFalse(os.path.isfile(meta_path))

    def test_wipe_keep_buckets(self):
        bucket_a_obj = {"key_a": "val_a"}
        bucket_b_obj = {"key_b": "valb_"}
        self.storage.create_bucket("bucket_a", bucket_a_obj)
        self.storage.create_bucket("bucket_b", bucket_b_obj)
        self.storage.wipe(keep_buckets=True)
        meta_path = _get_meta_path()

        with open(meta_path, "r") as file:
            meta = json.load(file)
            self.assertEqual(meta["buckets"]["bucket_a"], bucket_a_obj)
            self.assertEqual(meta["buckets"]["bucket_b"], bucket_b_obj)
            self.assertEqual(meta["objects"], {})
            self.assertEqual(meta["resumable"], {})



import datetime
import json
import logging

import fs
from fs.errors import FileExpected, ResourceNotFound

from gcloud_storage_emulator.exceptions import Conflict, NotFound
from gcloud_storage_emulator.settings import STORAGE_BASE, STORAGE_DIR

logger = logging.getLogger(__name__)


class Storage(object):
    def __init__(self, use_memory_fs=False):
        self._use_memory_fs = use_memory_fs
        self._pwd = fs.open_fs(self.get_storage_base())
        try:
            self._fs = self._pwd.makedir(STORAGE_DIR)
        except fs.errors.DirectoryExists:
            self._fs = self._pwd.opendir(STORAGE_DIR)

        self._read_config_from_file()

    def _write_config_to_file(self):
        data = {
            "buckets": self.buckets,
            "objects": self.objects,
            "resumable": self.resumable,
        }

        with self._fs.open(".meta", mode="w") as meta:
            json.dump(data, meta, indent=2)

    def _read_config_from_file(self):
        try:
            with self._fs.open(".meta", mode="r") as meta:
                data = json.load(meta)
                self.buckets = data.get("buckets")
                self.objects = data.get("objects")
                self.resumable = data.get("resumable")
        except ResourceNotFound:
            self.buckets = {}
            self.objects = {}
            self.resumable = {}

    def _get_or_create_dir(self, bucket_name, file_name):
        try:
            bucket_dir = self._fs.makedir(bucket_name)
        except fs.errors.DirectoryExists:
            bucket_dir = self._fs.opendir(bucket_name)

        dir_name = fs.path.dirname(file_name)
        return bucket_dir.makedirs(dir_name, recreate=True)

    def get_storage_base(self):
        """Returns the pyfilesystem-compatible fs path to the storage

        This is the OSFS if using disk storage, or "mem://" otherwise.
        See https://docs.pyfilesystem.org/en/latest/guide.html#opening-filesystems for more info

        Returns:
            string -- The relevant filesystm
        """

        if self._use_memory_fs:
            return "mem://"
        else:
            return STORAGE_BASE

    def get_bucket(self, bucket_name):
        """Get the bucket resourec object given the bucket name

        Arguments:
            bucket_name {str} -- Name of the bucket

        Returns:
            dict -- GCS-like Bucket resource
        """

        return self.buckets.get(bucket_name)

    def get_file_list(self, bucket_name, prefix=None, delimiter=None):
        """Lists all the blobs in the bucket that begin with the prefix.

        This can be used to list all blobs in a "folder", e.g. "public/".

        The delimiter argument can be used to restrict the results to only the
        "files" in the given "folder". Without the delimiter, the entire tree under
        the prefix is returned. For example, given these blobs:

            a/1.txt
            a/b/2.txt

        If you just specify prefix = 'a', you'll get back:

            a/1.txt
            a/b/2.txt

        However, if you specify prefix='a' and delimiter='/', you'll get back:

            a/1.txt

        Additionally, the same request will return blobs.prefixes populated with:

            a/b/

        Source: https://cloud.google.com/storage/docs/listing-objects#storage-list-objects-python
        """

        if bucket_name not in self.buckets:
            raise NotFound

        bucket_objects = self.objects.get(bucket_name, {})
        if prefix:
            # TODO: Still need to implement the last part of the doc string above to
            # TODO: populate blobs.prefixes when using a delimiter.
            return list(file_object for file_name, file_object in bucket_objects.items()
                        if file_name.startswith(prefix)
                        and (not delimiter or delimiter not in file_name[len(prefix+delimiter):]))
        else:
            return list(bucket_objects.values())

    def create_bucket(self, bucket_name, bucket_obj):
        """Create a bucket object representation and save it to the current fs

        Arguments:
            bucket_name {str} -- Name of the GCS bucket
            bucket_obj {dict} -- GCS-like Bucket resource

        Returns:
            [type] -- [description]
        """

        self.buckets[bucket_name] = bucket_obj
        self._write_config_to_file()
        return bucket_obj

    def create_file(self, bucket_name, file_name, content, file_obj):
        """Create a text file given a string content

        Arguments:
            bucket_name {str} -- Name of the bucket to save to
            file_name {str} -- File name used to store data
            content {bytes} -- Content of the file to write
            file_obj {dict} -- GCS-like Object resource
        """

        file_dir = self._get_or_create_dir(bucket_name, file_name)

        base_name = fs.path.basename(file_name)
        with file_dir.open(base_name, mode="wb") as file:
            file.write(content)
            bucket_objects = self.objects.get(bucket_name, {})
            bucket_objects[file_name] = file_obj
            self.objects[bucket_name] = bucket_objects
            self._write_config_to_file()

    def create_resumable_upload(self, bucket_name, file_name, file_obj):
        """Initiate the necessary data to support partial upload.

        This doesn't fully support partial upload, but expect the secondary PUT
        call to send all the data in one go.

        Basically, we try to comply to the bare minimum to the API described in
        https://cloud.google.com/storage/docs/performing-resumable-uploads ignoring
        any potential network failures

        Arguments:
            bucket_name {string} -- Name of the bucket to save to
            file_name {string} -- File name used to store data
            file_obj {dict} -- GCS Object resource

        Returns:
            str -- id of the resumable upload session (`upload_id`)
        """

        file_id = "{}:{}:{}".format(bucket_name, file_name, datetime.datetime.now())
        self.resumable[file_id] = file_obj
        self._write_config_to_file()
        return file_id

    def create_file_for_resumable_upload(self, file_id, content):
        """Create a binary file following a partial upload request

        This also updates the meta with the final file-size

        Arguments:
            file_id {str} -- the `upload_id` of the partial upload session
            content {bytes} -- raw content to add to the file

        Returns:
            dict -- GCS-like Object resource
        """

        file_obj = self.resumable[file_id]
        bucket_name = file_obj["bucket"]
        file_name = file_obj["name"]
        file_dir = self._get_or_create_dir(bucket_name, file_name)

        base_name = fs.path.basename(file_name)
        with file_dir.open(base_name, mode="wb") as file:
            file.write(content)

        file_obj["size"] = str(len(content))
        bucket_objects = self.objects.get(bucket_name, {})
        bucket_objects[file_name] = file_obj
        self.objects[bucket_name] = bucket_objects
        del self.resumable[file_id]
        self._write_config_to_file()

        return file_obj

    def get_file_obj(self, bucket_name, file_name):
        """Gets the meta information for a file within a bucket

        Arguments:
            bucket_name {str} -- Name of the bucket
            file_name {str} -- File name

        Raises:
            NotFound: Raised when the object doesn't exist

        Returns:
            dict -- GCS-like Object resource
        """

        try:
            return self.objects[bucket_name][file_name]
        except KeyError:
            raise NotFound

    def get_file(self, bucket_name, file_name):
        """Get the raw data of a file within a bucket

        Arguments:
            bucket_name {str} -- Name of the bucket
            file_name {str} -- File name

        Raises:
            NotFound: Raised when the object doesn't exist

        Returns:
            bytes -- Raw content of the file
        """

        try:
            bucket_dir = self._fs.opendir(bucket_name)
            return bucket_dir.open(file_name, mode="rb").read()
        except (FileExpected, ResourceNotFound) as e:
            logger.error("Resource not found:")
            logger.error(e)
            raise NotFound

    def delete_bucket(self, bucket_name):
        """Delete a bucket's meta and file

        Arguments:
            bucket_name {str} -- GCS bucket name

        Raises:
            NotFound: If the bucket doesn't exist
            Conflict: If the bucket is not empty or there are pending uploads
        """
        bucket_meta = self.buckets.get(bucket_name)
        if bucket_meta is None:
            raise NotFound("Bucket with name '{}' does not exist".format(bucket_name))

        bucket_objects = self.objects.get(bucket_name, {})

        if len(bucket_objects.keys()) != 0:
            raise Conflict("Bucket '{}' is not empty".format(bucket_name))

        resumable_ids = [
            file_id
            for (file_id, file_obj) in self.resumable.items()
            if file_obj.get('bucket') == bucket_name
        ]

        if len(resumable_ids) != 0:
            raise Conflict("Bucket '{}' has pending upload sessions".format(bucket_name))

        del self.buckets[bucket_name]

        self._delete_dir(bucket_name)
        self._write_config_to_file()

    def delete_file(self, bucket_name, file_name):
        try:
            self.objects[bucket_name][file_name]
        except KeyError:
            raise NotFound("Object with name '{}' does not exist in bucket '{}'".format(bucket_name, file_name))

        del self.objects[bucket_name][file_name]

        self._delete_file(bucket_name, file_name)
        self._write_config_to_file()

    def _delete_file(self, bucket_name, file_name):
        try:
            with self._fs.opendir(bucket_name) as bucket_dir:
                bucket_dir.remove(file_name)
        except ResourceNotFound:
            logger.info("No file to remove '{}/{}'".format(bucket_name, file_name))

    def _delete_dir(self, path, force=True):
        try:
            remover = self._fs.removetree if force else self._fs.removedir
            remover(path)
        except ResourceNotFound:
            logger.info("No folder to remove '{}'".format(path))

    def wipe(self, keep_buckets=False):
        existing_buckets = self.buckets
        self.buckets = {}
        self.objects = {}
        self.resumable = {}

        try:
            self._fs.remove('.meta')
            for path in self._fs.listdir('.'):
                self._fs.removetree(path)
        except ResourceNotFound as e:
            logger.warning(e)

        if keep_buckets:
            for k, v in existing_buckets.items():
                self.create_bucket(k, v)



#!/usr/bin/env python

import argparse
import logging
import sys

from gcloud_storage_emulator.handlers.buckets import create_bucket
from gcloud_storage_emulator.server import create_server
from gcloud_storage_emulator.storage import Storage

# One after gcloud-task-emulator one
DEFAULT_PORT = 9023
DEFAULT_HOST = "localhost"


def run_server(host, port, memory=False, default_bucket=None):
    server = create_server(host, port, memory, default_bucket)
    return server.run()


def wipe(keep_buckets=False):
    print("Wiping...")
    server = create_server(None, None, False)
    server.wipe(keep_buckets=keep_buckets)
    print("Done.")
    return 0


def prepare_args_parser():
    parser = argparse.ArgumentParser(description="Google Cloud Storage Emulator")
    subparsers = parser.add_subparsers(title="subcommands", dest="subcommand")

    start = subparsers.add_parser("start", help="start the emulator")
    start.add_argument(
        "--port", type=int, help="the port to run the server on", default=DEFAULT_PORT
    )
    start.add_argument(
        "-H", "--host", help="the host to run the server on", default=DEFAULT_HOST
    )
    start.add_argument(
        "--default-bucket", help="The default bucket. If provided, bucket will be created automatically"
    )
    start.add_argument("-q", "--quiet", action="store_true", default=False, help="only outputs critical level logging")
    start.add_argument("-M", "--no-store-on-disk", action="store_true", default=False, help="use in-memory storage")

    wipe = subparsers.add_parser("wipe", help="Wipe the local data")
    wipe.add_argument("--keep-buckets", help="If provided the data will be wiped but the existing buckets are kept")

    create_bucket = subparsers.add_parser("create_bucket", help="create bucket")
    create_bucket.add_argument(
        "-n", "--name",
        help="Name of the new bucket"
    )

    return parser, subparsers


def main():
    parser, subparsers = prepare_args_parser()
    args = parser.parse_args()
    if args.subcommand not in subparsers.choices.keys():
        parser.print_usage()
        sys.exit(1)

    if args.subcommand == "wipe":
        answer = input("This operation will IRREVERSIBLY DELETE all your data. Do you wish to proceed? [y/N] ").lower()
        if answer in ("y", "ye", "yes"):
            sys.exit(wipe(keep_buckets=args.keep_buckets))
        else:
            print("wipe command cancelled")
            sys.exit(1)

    if args.subcommand == "create_bucket":
        storage = Storage()
        create_bucket(args.name, storage)
        sys.exit(1)

    root = logging.getLogger("")
    stream_handler = logging.StreamHandler()
    root.addHandler(stream_handler)
    if args.quiet:
        root.setLevel(logging.CRITICAL)
    else:
        root.setLevel(logging.DEBUG)
    sys.exit(run_server(args.host, args.port, args.no_store_on_disk, args.default_bucket))


if __name__ == "__main__":
    main()
